/*========================================================================
     This is a part of "gpop" - Gaussian output post processor
             Copyright (c) 2002-2009 by A. Miyoshi, Univ. Tokyo
========================================================================*/

#ifndef _GPOPLIBC_H_
#define _GPOPLIBC_H_

#include "gpoplibf.h"

/*------------------------------------------------------------------------
     Tabulated informations on elements and methods
------------------------------------------------------------------------*/

class elemInfo {
 public:                             // --- data members ---
  int aNbr;
  string aSym;
  double aWgt;
};

class tableOfElements : public vector<elemInfo> {
 public:                         // --- member functions ---
  tableOfElements();
  int getAtmcNbr(string &sm);
  double getAtmcWeight(int anbr);
  void setAtmcWeight(int anbr, double wgt);
  string getAtmcSym(int anbr);
};                                        // --- globals ---
void reformAtomicSymbol(string &s);

class methodInfo {
 public:                             // --- data members ---
  string method, basis;
  double scaleFactVib, scaleFactZPE, scaleFactGT;
};

class tableOfMethods : public vector<methodInfo> {
 public:                         // --- member functions ---
  tableOfMethods();
  int getIdx(string &rt);
};

/*------------------------------------------------------------------------
     Generic Vector
------------------------------------------------------------------------*/

class gnVector : public vector<double> {
 public:                         // --- member functions ---
  gnVector();
  gnVector(int siz);
  void init(int siz);
  double operator*(gnVector gv);        // inner product
};

/*------------------------------------------------------------------------
     Generic Matrix
------------------------------------------------------------------------*/

class gnMatrix : public vector<gnVector> {
 public:                         // --- member functions ---
  gnMatrix();
  gnMatrix(int rsiz, int csiz);
  void init(int rsiz, int csiz);
  gnVector operator*(gnVector gv);
  gnMatrix operator~();                 // transposition
  gnMatrix sqElement();                 // each element squared
};

/*------------------------------------------------------------------------
     Cartesian coordinate (also a vector)
------------------------------------------------------------------------*/

class coordCart : public vector<double> {
 public:                         // --- member functions ---
  coordCart();
  double &x();
  double &y();
  double &z();
  void fillZero();
  double vctLength();
  void normalize();
  coordCart invert();
  coordCart operator-();
  coordCart add(coordCart crd);
  coordCart operator+(coordCart crd);
  coordCart addMe(coordCart crd);
  coordCart operator+=(coordCart crd);
  coordCart subtract(coordCart crd);
  coordCart operator-(coordCart crd);
  coordCart subtractMe(coordCart crd);
  coordCart operator-=(coordCart crd);
  coordCart multiply(double fac);
  coordCart operator*(double fac);
  coordCart multiplyMe(double fac);
  coordCart operator*=(double fac);
  coordCart divideMe(double den);
  coordCart operator/=(double den);
  double innerProduct(coordCart crd);
  double operator*(coordCart crd);
  coordCart outerProduct(coordCart crd);
  double dirCosine(coordCart crd);
  double angleWith(coordCart crd);
  int minAbsIndex();
  ostream &print(ostream &os);
};                                        // --- globals ---
ostream &operator<<(ostream &os, coordCart &crd);
coordCart operator*(double fac, coordCart crd);
coordCart vectProd(coordCart crdA, coordCart crdB);
int cycIncCrdIdx(int idx);
int cycDecCrdIdx(int idx);
coordCart vertPnt(coordCart &a, coordCart &b, coordCart &c);
double vertDist(coordCart &a, coordCart &b, coordCart &c);

/*------------------------------------------------------------------------
     Cartesian axis
------------------------------------------------------------------------*/

class axisCart {
 public:                             // --- data members ---
  coordCart origin;      // origin coordinate
  coordCart direct;      // direction vector
 public:                         // --- member functions ---
  void setByPoints(coordCart &pnt0, coordCart &pnt1);
  void setByOrgVect(coordCart &org, coordCart &vct);
  coordCart vertPoint(coordCart &pntC);
  double distance(coordCart &pntC);
};

/*------------------------------------------------------------------------
     Cartesian matrix
------------------------------------------------------------------------*/

class matrxCart : public vector<coordCart> {
 public:                         // --- member functions ---
  matrxCart();
  coordCart &X();
  coordCart &Y();
  coordCart &Z();
  double &X(int idxc);
  double &Y(int idxc);
  double &Z(int idxc);
  coordCart multVect(coordCart crd);
  coordCart operator*(coordCart crd);
  double determinant();
};

/*------------------------------------------------------------------------
     Polar coordinate
------------------------------------------------------------------------*/

class coordPolar {
 public:                             // --- data members ---
  double r, th, ph;
 public:                         // --- member functions ---
  coordPolar();
  void setFrom(coordCart cc);
  double degTh();
  double degPh();
  coordPolar subtract(coordPolar cp);
  coordPolar operator-(coordPolar cp);
};

/*------------------------------------------------------------------------
     Gaussian Job Informations
------------------------------------------------------------------------*/

class jobInfos {
 public:                             // --- data members ---
  string version, revision, comment;
  int methodIdx;
  string method, basis;
  double scaleFactVib, scaleFactZPE, scaleFactGT;
  string SCFenergy;
 public:                         // --- member functions ---
  jobInfos();
  void clear();
  int setFrom(string &gauv, string &route, string &comm, string &scfe);
  int restoreFrom(istream &is);
  void setMethodProp();
  ostream &print(ostream &os);
};                                        // --- globals ---
ostream &operator<<(ostream &os, jobInfos &ji);

/*------------------------------------------------------------------------
     Base classes for geometry and frequency
------------------------------------------------------------------------*/

class devnInfo : public vector<coordCart> {
 public:                         // --- member functions ---
  devnInfo add(devnInfo di);
  devnInfo operator+(devnInfo di);
  devnInfo addMe(devnInfo di);
  devnInfo operator+=(devnInfo di);
  devnInfo subtract(devnInfo di);
  devnInfo operator-(devnInfo di);
  devnInfo subtractMe(coordCart crd);
  devnInfo operator-=(coordCart crd);
  devnInfo subtractMe(devnInfo di);
  devnInfo operator-=(devnInfo di);
  devnInfo multiply(double fac);
  devnInfo operator*(double fac);
  devnInfo divideMe(double den);
  double sumInnerProd(devnInfo di);
  double operator*(devnInfo di);
  double amplitude();
  devnInfo normalize();
};                                        // --- globals ---
devnInfo operator*(double fac, devnInfo di);

class atomCoord : public coordCart {
 public:                             // --- data members ---
  int atomicNumber;
  int atomicType;
 public:                         // --- member functions ---
  atomCoord();
  atomCoord setCoord(coordCart crd);
  atomCoord operator=(coordCart crd);
  atomCoord add(coordCart crd);
  atomCoord operator+(coordCart crd);
  atomCoord subtractMe(coordCart crd);
  atomCoord operator-=(coordCart crd);
};

/*------------------------------------------------------------------------
     Geometry
------------------------------------------------------------------------*/

class geomInfo : public vector<atomCoord> {
 public:                             // --- data members ---
  coordCart com;
  matrxCart prcAxes;
  double momInert[3];
  double molW;
 public:                         // --- member functions ---
  int restoreFrom(istream &is);
  void outFWformat(ostream &os);
  void setMolW();
  void setCentOfMass();
  void setPrcAxesMomInert();
  matrxCart getTensorInert();
  geomInfo add(devnInfo di);
  geomInfo operator+(devnInfo di);
  geomInfo subtractMe(coordCart crd);
  geomInfo operator-=(coordCart crd);
  void transformMe(matrxCart mc);
  void beCanonical();
  coordCart getVelCOM(devnInfo &di);
  double getAngVPrcAx(devnInfo &di, int iax);
  devnInfo getUnitRotPrcAx(int iax);
  ostream &print(ostream &os);
};                                        // --- globals ---
ostream &operator<<(ostream &os, geomInfo &gi);

class molecGeom {
 public:                             // --- data members ---
  int nAtms, gElec, nVibm, isLinear;
  string pntGrp, elState;
  double rotSymNbr, numIsomers, rotcGHz[3];
  double molW;
  map<int, int> chemform;
  geomInfo geom;
 public:                         // --- member functions ---
  int setFrom(vector<string> &gm, string &estate, string &chgmul);
  void clear();
  int restoreFrom(istream &is);
  ostream &print(ostream &os);
};                                        // --- globals ---
ostream &operator<<(ostream &os, molecGeom &mg);

/*------------------------------------------------------------------------
     Frequencies
------------------------------------------------------------------------*/

class freqInfo {
 public:                             // --- data members ---
  string sym;
  double freq, rmass, fconst, irint, ramact, depol;
  devnInfo nrm;
};

class vibSimil {
 public:                             // --- data members ---
  int idx;
  double simil, freq;
};

class vibSimList : public vector<vibSimil> {
};

class freqTable : public vector<freqInfo> {
 public:                             // --- data members ---
  double zpe_Unscaled, zpe, zpe_GT;
 public:                         // --- member functions ---
  int setFrom(vector<string> &fr, jobInfos &ji, molecGeom &mg);
  int restoreFrom(istream &is);
  ostream &print(ostream &os);
  int countImag();
  vector<pair<int, double> > getMostAlike(devnInfo &di);
  vibSimList findSimil(devnInfo &di);
};                                        // --- globals ---
ostream &operator<<(ostream &os, freqTable &ft);

/*------------------------------------------------------------------------
     Molecular moiety (for internal rotor calculations)
------------------------------------------------------------------------*/

class molMoiety {
 public:                             // --- data members ---
  vector<int> lstAtms;
  geomInfo geom;
  double appI, appB;        // I,B of rotor
  double symRedI, symRedB;  // reduced I,B of symmetric rotor
  double asmRedI, asmRedB;  // reduced I,B of asymmetric rotor
  coordCart topCOM, topOrigin;
  coordCart topAxes[3];
 public:                         // --- member functions ---
  void setTopProp(atomCoord &cPiv, molecGeom &mg);
  void setRotDevn(atomCoord &cPiv, devnInfo &rotDI);
 private:
  void setAppTopProp(atomCoord &cPiv);
  void defineTopAxes(atomCoord &cPiv);
  void setAsymTopProp(atomCoord &cPiv, molecGeom &mg);
};

/*------------------------------------------------------------------------
     Molecule consisting of two molMoiety's (internal rotor)
------------------------------------------------------------------------*/

class molByMoieties : public vector<molMoiety> {
 public:                             // --- data members ---
  int isWholeMolec;
  double dihedAngle;
  devnInfo intRotDevn;
 public:                         // --- member functions ---
  int setList(vector<string> &moin, molecGeom &mg);
  void setProp(molecGeom &mg);
  ostream &print(ostream &os);
  void setIntRotDevn(molecGeom &mg);
};                                        // --- globals ---
ostream &operator<<(ostream &os, molByMoieties &mbm);

/*------------------------------------------------------------------------
     global library functions
------------------------------------------------------------------------*/

// --- gaussian-log input functions --------------------------------------

int findGausJobHead(istream &inf, string &gauv);
string getGausRouteSect(istream &inf);
string getGausCommLine(istream &inf);
// int getGausArchive(istream &inf, string &arch);
int findGausJobEnd(istream &inf, string &jcpu);

// --- gpop file input / output function ---------------------------------

int readGpo(string &basfln, jobInfos &ji, molecGeom &mg, freqTable &ft);

 /*--- following functions return true or false ---*/
int getlGpoMatch(istream &inStrm, const string &ref);
int getlGpoStrv(istream &inStrm, string &vStr);
int getlGpoIntv(istream &inStrm, int &vInt);
int getlGpoDblv(istream &inStrm, double &vDbl);
int getlGpoDblvMlt(istream &inStrm, double mvDbl[], int nVals);
int getlGpoBlnv(istream &inStrm, int &vBln);
int getlGpoChmFrm(istream &inStrm, map<int, int> &cf);
int getlGpoAtmCoord(istream &inStrm, atomCoord &ac);
int getlAmbAtmCoord(istream &inStrm, atomCoord &ac);
int getlGpofrqInfo(istream &inStrm, freqInfo &fi);
int getlGpoCrdCart(istream &inStrm, coordCart &crd);
int getlGpoKeyToks(istream &inStrm, string &key, vector<string> &toks);

 /*--- this returns NORM or ERRE ---*/
int getLineGpop(istream &inStrm, string &sLine);

int getBlnvStr(string s);

int outFileGpo(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft, string &gaufln);
int outFileTst(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft);
string strGpoBlnv(int &vBln);

/*------------------------------------------------------------------------
     Declaration of globals
------------------------------------------------------------------------*/

extern tableOfElements elm;
extern tableOfMethods gmth;

#endif

