/* ======================================================================
     This is a part of "gpop" - Gaussian output post processor
             Copyright (c) 2002-2005 by A. Miyoshi, Univ. Tokyo
 ====================================================================== */

#ifndef _GPOPLIBT_H_
#define _GPOPLIBT_H_

#include "gpoplibr.h"

/*------------------------------------------------------------------------
     internal rotor description for modInfoTST
------------------------------------------------------------------------*/

/*----- an internal rotor ------------------------------------*/

class intRotor {
 public:                             // --- data members ---
  int idxVib;            // index of vib. to be hind. rot.
  int isAuto;            // auto-assignment flag
  double csim;           // similarity coefficient
  int nSym;              // potential symmetry number
  double redB;           // reduced rotational constant [cm-1]
  double V0;             // V0 [cm-1]
  vector<int> rotAtms;   // list of rotor atoms
  vector<int> cntAtms;   // list of counter part atoms
  string descript;       // description of the mode
  molByMoieties mbm;
 public:                         // --- member functions ---
  intRotor();
  void clear();
  int setFrom(int iv, int ns, vector<string> &mi, string &ds,
   molecGeom &mg, freqTable &ft);
  int restoreFrom(istream &is);
  ostream &print(ostream &os);
};                                        // --- globals ---
ostream &operator<<(ostream &os, intRotor &ir);

/*----- list of internal rotors ------------------------------*/

class listIntRotors : public vector<intRotor> {
 public:                         // --- member functions ---
  int restoreFrom(istream &is, int nir);
  ostream &print(ostream &os);
  int isIntRotor(int iv);
  int getIdxIntRotor(int iv);
  int findDupVibIdx(vector<int> &irdup, map<int, set<int> > &duplst);
  void diagnIntRotAssign(freqTable &ft);
};                                        // --- globals ---
ostream &operator<<(ostream &os, listIntRotors &lir);

/*------------------------------------------------------------------------
     TST related modification informations
------------------------------------------------------------------------*/

class modInfoTST {
 public:                             // --- data members ---
  int isTS;              // flag indicates TS
  double energyTST;      // energy for TST (hartree)
  listIntRotors hndRots;
  listStates eStates;
  listStates isomStates;
 public:                         // --- member functions ---
  modInfoTST();
  void clear();
  int restoreFrom(istream &is);
  ostream &print(ostream &os);
};                                        // --- globals ---
ostream &operator<<(ostream &os, modInfoTST &mit);

/*------------------------------------------------------------------------
     List of temperatures
------------------------------------------------------------------------*/

class tempList : public vector<double> {
 protected:
  enum errcode { FAILopen, EMPfile, INVkey, INVstep, INVrange, INVrngMANY,
                 INVparN, INVtemp };
 public:                         // --- member functions ---
  tempList();
  void setDefault();
  int readFrom(string flnam);
  int addFrom(string slin);
  int setBy(string slin);
  void sortElimDup();
  int addRange(double st, double et, double dt);
  int addRangeVS(vector<string> &vs);
  int setByRange(double st, double et, double dt);
  int setByRangeVS(vector<string> &vs);
  int addRecipRange(double numr, double stt, double lst, double stp);
  int addRecipRangeVS(vector<string> &vs);
  int setByRecipRange(double numr, double stt, double lst, double stp);
  int setByRecipRangeVS(vector<string> &vs);
  int addList(vector<double> &vd);
  int addListVS(vector<string> &vs);
  int setByList(vector<double> &vd);
  int setByListVS(vector<string> &vs);
  ostream &print(ostream &os);
  double maxTemp();
 protected:
  int errms(errcode erc, string sv = "");
};                                        // --- globals ---
ostream &operator<<(ostream &os, tempList &tl);

/*------------------------------------------------------------------------
     molecular internal mode
------------------------------------------------------------------------*/

class molIntMode {
 public:                             // --- data members ---
  enum { Vibrator, FreeRotor, HindRotor } type;
  int nSym;
  double freq, redB;     // [cm-1]
 public:                         // --- member functions ---
  molIntMode();
  void clear();
  ostream &print(ostream &os);
  double modeQ(double T);
  double modeU(double T);
  double modeS(double T);
  double modeCv(double T);
};                                        // --- globals ---
ostream &operator<<(ostream &os, molIntMode &minm);
 // --- global thermodynamic functions
                                          // - hind. rot. Pitzer-Gwinn
double qPGHindRot(double T, double freq, double B, int sgm);
  double uPGHindRot(double T, double freq, double B, int sgm);
  double sPGHindRot(double T, double freq, double B, int sgm);
  double cvPGHindRot(double T, double freq, double B, int sgm);
double qHOvib(double T, double freq);     // - vibrational HO/quantum
  double uHOvib(double T, double freq);
  double sHOvib(double T, double freq);
  double cvHOvib(double T, double freq);
double qClHOvib(double T, double freq);   // - vibrational HO/classical
  double uClHOvib(double T, double freq);
  double sClHOvib(double T, double freq);
  double cvClHOvib(double T, double freq);
                                          // - hindered rotor classical
double qClHindRot(double T, double B, int sgm, double V0);
  double uClHindRot(double T, double B, int sgm, double V0);
  double sClHindRot(double T, double B, int sgm, double V0);
  double cvClHindRot(double T, double B, int sgm, double V0);
                                          // - 1-D free rotor
double q1DRRrot(double T, double B, int sgm);
  double u1DRRrot(double T, double B, int sgm);
  double s1DRRrot(double T, double B, int sgm);
  double cv1DRRrot(double T, double B, int sgm);
double besselI(int n, double x);
double factorial(int n);

/*------------------------------------------------------------------------
     thermofunc calculation results
------------------------------------------------------------------------*/

/*----- Thermofunc calculation results at a temperature ----------------*/

class res1Therm {
 public:                             // --- data members ---
  double T, S, Cp, H;
 public:                         // --- member functions ---
  void printCSV(ostream &os);
  void printCSVsubtH(ostream &os, double Hr);
};

/*----- Thermofunc calculation for a molecule --------------------------*/

class resMolTherm : public vector<res1Therm> {
 public:                             // --- data members ---
  res1Therm ref;
 public:                         // --- member functions ---
  void printHeadCSV(ostream &os);
  void printHeadCSVs(ostream &os);
  void printCSVsubtH(ostream &os);
};

/*------------------------------------------------------------------------
     molecular properties for TST calculation
------------------------------------------------------------------------*/

class molPropTST {
 public:                             // --- data members ---
  enum { Invalid, Atom, Linear, NonLinear } molecType;
  string molnam;
  int rotSymNbr;
  double sigma_nIsom;
  double molW;           // [amu]
  double rotC[3];        // [cm-1]
  double energyTST;      // [hartree]
  double imagFreqTS;     // [cm-1]
  vector<molIntMode> intMode;
  listStates eStates;
  listStates isomStates;
 public:                         // --- member functions ---
  molPropTST();
  void clear();
  int readFromTST(string basfln);
  void setFrom(string &basfln, jobInfos &ji, molecGeom &mg, freqTable &ft,
   modInfoTST &mit);
  ostream &print(ostream &os);
                              // --- thermodynamics
  void getTherm(double T, double &S, double &Cp, double &H);
  resMolTherm getMolTherm(tempList &tl);
  res1Therm get1Therm(double T);
  double totalQ(double T);
    double totalU(double T);
    double totalS1bar(double T);
    double totalCv(double T);
  double transQ(double T);    // - translational
    double transU(double T);
    double transS1bar(double T);
    double transCv(double T);
  double elecQ(double T);     // - electronic
    double elecU(double T);
    double elecS(double T);
    double elecCv(double T);
  double isomQ(double T);     // - isomer-contribution
    double isomU(double T);
    double isomS(double T);
    double isomCv(double T);
  double rotQ(double T);      // - extern-rotational
    double rotU(double T);
    double rotS(double T);
    double rotCv(double T);
                              // - internal mode (vib/FR/HR)
  double imodQ(double T, double &qV, double &qFR, double &qHR);
    double imodU(double T, double &uV, double &uFR, double &uHR);
    double imodS(double T, double &sV, double &sFR, double &sHR);
    double imodCv(double T, double &cvV, double &cvFR, double &cvHR);
};                                        // --- globals ---
ostream &operator<<(ostream &os, molPropTST &mpt);

/*----- molecular information set for reaction specification -----------*/

class molInfoSet {
 public:                             // --- data members ---
  enum { Invalid = 0, Energy = 1, File = 2, ImagFreq = 4 };
  int status;
  int isTS;
  double energy;
  double imagFreqTS;
  vector<string> basnams;
  vector<molPropTST> molp;
 public:                         // --- member functions ---
  molInfoSet();
  void clear();
  int readFrom(istream &is);
  int readProp(string ftyp);
  ostream &print(ostream &os);
  double molsetQ(double T);
  double molsetU(double T);
  double molsetS1bar(double T);
  double molsetCv(double T);
  double molsetH(double T);
  double molsetG1bar(double T);
};                                        // --- globals ---
ostream &operator<<(ostream &os, molInfoSet &mis);

/*----- TST calculation results at a temperature -----------------------*/

class resultTST {
 public:                             // --- data members ---
  double T;
  double kTST, Keq;
  double corTunSh, corTunEck;
 public:                         // --- member functions ---
  resultTST();
  void clear();
  ostream &print(ostream &os);
};                                        // --- globals ---
ostream &operator<<(ostream &os, resultTST &rT);

/*----- tunneling correction -------------------------------------------*/

class tunnelCorr {
 public:                             // --- data members ---
  int flgShavitt;
  int flgEckart;
  double vMax, vDiff, fImg;             // input asym. Eckart parameters
  double vB, beta, tpcoshD;             // derived asym. Eckart parameters
  double eneMin, eneMax, eneGrain;      // integration parameters
  double maxNumError;
 public:                         // --- member functions ---
  tunnelCorr();
  void clear();
  void initShavitt(double v0, double fi);
  void initEckart(double v0, double vD, double fi, double maxT);
  ostream &print(ostream &os);
  double corrFacShavitt(double T);
  double corrFacEckart(double T);
 private:
  double fClAnalytic(double T);
  double fClNumeric(double T);
  double fQmNumeric(double T);
  double tpEck(double E);
};
ostream &operator<<(ostream &os, tunnelCorr &tC);

/*----- reaction information for TST calculation -----------------------*/

class reactionInfo {
 public:                             // --- data members ---
  enum { Invalid = 0, Energy = 1, File = 2, ImagFreq = 4 };
  int calRat, calTun, calEck, calEql;
  string fileType;
  string baseName;
  double enCnvFac;       // energy conversion factor
  tempList temperat;
  molInfoSet mReac, mTS, mProd;
  tunnelCorr tnlC;
  vector<resultTST> calcRes;
  double dU0, dH298, dS298, dUts0, dHts298, dSts298;
  double apA[3], apER[3], fmxErr[3];
 public:                         // --- member functions ---
  reactionInfo();
  void clear();
  int readFrom(string basfln);
  int readProp();
  ostream &print(ostream &os);
  void calculate();
  void fitArrhenius();
  int outfileCSV();
};                                        // --- globals ---
ostream &operator<<(ostream &os, reactionInfo &ri);

/*------------------------------------------------------------------------
     global library functions
------------------------------------------------------------------------*/

int readGpo(string &basfln, jobInfos &ji, molecGeom &mg, freqTable &ft,
 modInfoTST &mit);
int outFileGpo(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft, modInfoTST &mit);
int outFileTst(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft, modInfoTST &mit);

#endif

